@@ -20,12 +20,15 @@
  *
  */
 
-
+#ifdef HAVE_CONFIG_H
 #include <config.h>
+#endif
+
 #include <stdio.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
+#include <unistd.h>
 #include <string.h>
 #include <ctype.h>
 #include "gmime-part.h"
@@ -115,6 +118,8 @@ g_mime_part_destroy (GMimePart *mime_part)
 			
 			g_list_free (mime_part->disposition->params);
 		}
+		
+		g_free (mime_part->disposition);
 	}
 	
 	if (mime_part->children) {
@@ -540,8 +545,9 @@ g_mime_part_get_content_disposition_parameter (GMimePart *mime_part, const gchar
 	GMimeParam *param;
 	
 	g_return_val_if_fail (mime_part != NULL, NULL);
-	g_return_val_if_fail (mime_part->disposition != NULL, NULL);
-	g_return_val_if_fail (mime_part->disposition->param_hash != NULL, NULL);
+	
+	if (!mime_part->disposition || !mime_part->disposition->param_hash)
+		return NULL;
 	
 	param = g_hash_table_lookup (mime_part->disposition->param_hash, name);
 	
@@ -597,15 +603,14 @@ g_mime_part_set_filename (GMimePart *mime_part, const gchar *filename)
  * and if not then checks the "name" parameter in the Content-Type.
  **/
 const gchar *
-g_mime_part_get_filename (GMimePart *mime_part)
+g_mime_part_get_filename (const GMimePart *mime_part)
 {
-	GMimeParam *param;
+	GMimeParam *param = NULL;
 	
 	g_return_val_if_fail (mime_part != NULL, NULL);
-	g_return_val_if_fail (mime_part->disposition != NULL, NULL);
-	g_return_val_if_fail (mime_part->disposition->param_hash != NULL, NULL);
 	
-	param = g_hash_table_lookup (mime_part->disposition->param_hash, "filename");
+	if (mime_part->disposition && mime_part->disposition->param_hash)
+		param = g_hash_table_lookup (mime_part->disposition->param_hash, "filename");
 	
 	if (!param || !param->value) {
 		/* check the "name" param in the content-type */
@@ -703,6 +708,77 @@ g_mime_part_set_content (GMimePart *mime_part, const gchar *content, guint len)
 }
 
 
+/**
+ * g_mime_part_set_content_byte_array: Set the content of the mime part
+ * @mime_part: Mime part
+ * @content: raw mime part content.
+ *
+ * Sets the content of the Mime Part (only non-multiparts)
+ **/
+void
+g_mime_part_set_content_byte_array (GMimePart *mime_part, GByteArray *content)
+{
+	g_return_if_fail (mime_part != NULL);
+	
+	if (mime_part->content)
+		g_byte_array_free (mime_part->content, TRUE);
+	
+	mime_part->content = content;
+}
+
+
+/**
+ * g_mime_part_append_pre_encoded_content:
+ * @mime_part: Mime part
+ * @content: content to append
+ * @len: content length
+ * @encoding: encoding type
+ *
+ * Append some pre-encoded content
+ **/
+void
+g_mime_part_append_pre_encoded_content (GMimePart *mime_part, const gchar *content,
+					guint len, GMimePartEncodingType encoding)
+{
+	gchar *raw;
+	guint clen;
+	
+	g_return_if_fail (mime_part != NULL);
+	g_return_if_fail (content != NULL);
+	
+	/* make sure we've got a content byte array */
+	if (mime_part->content == NULL)
+		mime_part->content = g_byte_array_new ();
+	
+	/* make sure we've got enough room for the new decoded data */
+	clen = mime_part->content->len;
+	g_byte_array_set_size (mime_part->content, clen + len);
+	
+	/* write the decoded data */
+	raw = mime_part->content->data + clen;
+	switch (encoding) {
+	case GMIME_PART_ENCODING_BASE64:
+		len = g_mime_utils_base64_decode_step (content, len, raw, &mime_part->append_state,
+						       &mime_part->append_save);
+		g_byte_array_set_size (mime_part->content, clen + len);
+		break;
+	case GMIME_PART_ENCODING_QUOTEDPRINTABLE:
+		len = g_mime_utils_quoted_decode_step (content, len, raw, &mime_part->append_state,
+						       &mime_part->append_save);
+		g_byte_array_set_size (mime_part->content, clen + len);
+		break;
+	default:
+		memcpy (raw, content, len);
+		
+		/* do some smart 8bit detection */
+		if (encoding == GMIME_PART_ENCODING_DEFAULT && g_mime_utils_text_is_8bit (raw, len))
+			encoding = GMIME_PART_ENCODING_8BIT;
+	}
+	
+	mime_part->encoding = encoding;
+}
+
+
 /**
  * g_mime_part_set_pre_encoded_content: Set the pre-encoded content of the mime part
  * @mime_part: Mime part
@@ -713,7 +789,8 @@ g_mime_part_set_content (GMimePart *mime_part, const gchar *content, guint len)
  * Sets the encoding type and raw content on the mime part after decoding the content.
  **/
 void
-g_mime_part_set_pre_encoded_content (GMimePart *mime_part, const gchar *content, guint len, GMimePartEncodingType encoding)
+g_mime_part_set_pre_encoded_content (GMimePart *mime_part, const gchar *content,
+				     guint len, GMimePartEncodingType encoding)
 {
 	gchar *raw;
 	gint save = 0, state = 0;
@@ -740,7 +817,7 @@ g_mime_part_set_pre_encoded_content (GMimePart *mime_part, const gchar *content,
 		memcpy (raw, content, len);
 		
 		/* do some smart 8bit detection */
-		if (encoding == GMIME_PART_ENCODING_DEFAULT && g_mime_utils_text_is_8bit (raw))
+		if (encoding == GMIME_PART_ENCODING_DEFAULT && g_mime_utils_text_is_8bit (raw, len))
 			encoding = GMIME_PART_ENCODING_8BIT;
 	}
 	
@@ -760,7 +837,9 @@ const gchar *
 g_mime_part_get_content (const GMimePart *mime_part, guint *len)
 {
 	g_return_val_if_fail (mime_part != NULL, NULL);
-	g_return_val_if_fail (mime_part->content != NULL, NULL);
+	
+	if (!mime_part->content)
+		return NULL;
 	
 	*len = mime_part->content->len;
 	
@@ -796,7 +875,8 @@ get_content_disposition (GMimePart *mime_part)
 	GList *params;
 	gchar *str;
 	
-	g_return_val_if_fail (mime_part->disposition != NULL, NULL);
+	if (!mime_part->disposition)
+		return NULL;
 	
 	params = mime_part->disposition->params;
 	
@@ -905,25 +985,25 @@ get_content (GMimePart *part)
 	return content;
 }
 
+
 /**
- * g_mime_part_to_string: Write the MIME Part to a string
+ * g_mime_part_write_to_string: Write the MIME Part to a string
  * @mime_part: MIME Part
  * @toplevel: mime part is the root mime part
+ * @string: output string
  *
- * Returns an allocated string containing the MIME Part. If toplevel
- * is set to TRUE, then the MIME Part header will contain needed MIME
- * headers for rfc822 messages.
+ * Writes the contents of the MIME Part to @string. If toplevel is set
+ * to TRUE, then the MIME Part header will contain needed MIME headers
+ * for rfc822 messages.
  **/
-gchar *
-g_mime_part_to_string (GMimePart *mime_part, gboolean toplevel)
+void
+g_mime_part_write_to_string (GMimePart *mime_part, gboolean toplevel, GString *string)
 {
-	gchar *str;
-	
-	g_return_val_if_fail (mime_part != NULL, NULL);
+	g_return_if_fail (mime_part != NULL);
+	g_return_if_fail (string != NULL);
 	
 	if (g_mime_content_type_is_type (mime_part->mime_type, "multipart", "*")) {
 		const gchar *boundary;
-		GString *contents;
 		GList *child;
 		gchar *content_type;
 		
@@ -936,43 +1016,27 @@ g_mime_part_to_string (GMimePart *mime_part, gboolean toplevel)
 		
 		content_type = get_content_type (mime_part->mime_type);
 		
-		contents = g_string_new ("");
+		if (toplevel) {
+			g_string_sprintfa (string, "MIME-Version: 1.0\n"
+					   "%s\n"  /* content-type */
+					   "This is a multi-part message in MIME format.\n\n",
+					   content_type);
+		} else {
+			g_string_sprintfa (string, "%s\n", content_type);
+		}
+		
+		g_free (content_type);
 		
 		child = mime_part->children;
 		while (child) {
-			gchar *part, *mime_string;
-			
-			mime_string = g_mime_part_to_string (child->data, FALSE);
-			
-			part = g_strdup_printf ("--%s\n%s\n", boundary, mime_string);
-			
-			g_string_append (contents, part);
-			
-			g_free (mime_string);
-			g_free (part);
+			g_string_sprintfa (string, "--%s\n", boundary);
+			g_mime_part_write_to_string (child->data, FALSE, string);
+			g_string_append_c (string, '\n');
 			
 			child = child->next;
 		}
 		
-		if (toplevel) {
-			/* add message header stuff first */
-			str = g_strdup_printf ("MIME-Version: 1.0\n"
-					       "%s\n"  /* content-type */
-					       "This is a multi-part message in MIME format.\n\n"
-					       "%s\n--%s--\n",
-					       content_type,
-					       contents->str,
-					       boundary);
-		} else {
-			str = g_strdup_printf ("%s\n"  /* content-type */
-					       "%s\n--%s--\n",
-					       content_type,
-					       contents->str,
-					       boundary);
-		}
-		
-		g_free (content_type);
-		g_string_free (contents, TRUE);
+		g_string_sprintfa (string, "\n--%s--\n", boundary);
 	} else {
 		gchar *content_type;
 		gchar *disposition;
@@ -982,9 +1046,6 @@ g_mime_part_to_string (GMimePart *mime_part, gboolean toplevel)
 		gchar *content_location;
 		gchar *content;
 		gchar *text;
-		GString *string;
-		
-		string = g_string_new ("");
 		
 		if (toplevel)
 			g_string_append (string, "MIME-Version: 1.0\n");
@@ -1027,7 +1088,7 @@ g_mime_part_to_string (GMimePart *mime_part, gboolean toplevel)
 		
 		/* Content-Location: */
 		if (mime_part->content_location) {
-			content_md5 = g_strdup_printf ("Content-Location: %s\n", mime_part->content_location);
+			content_location = g_strdup_printf ("Content-Location: %s\n", mime_part->content_location);
 			g_string_append (string, content_location);
 			g_free (content_location);
 		}
@@ -1054,10 +1115,31 @@ g_mime_part_to_string (GMimePart *mime_part, gboolean toplevel)
 		g_string_append_c (string, '\n');
 		
 		g_free (content);
-		
-		str = string->str;
-		g_string_free (string, FALSE);
 	}
+}
+
+
+/**
+ * g_mime_part_to_string: Write the MIME Part to a string
+ * @mime_part: MIME Part
+ * @toplevel: mime part is the root mime part
+ *
+ * Returns an allocated string containing the MIME Part. If toplevel
+ * is set to TRUE, then the MIME Part header will contain needed MIME
+ * headers for rfc822 messages.
+ **/
+gchar *
+g_mime_part_to_string (GMimePart *mime_part, gboolean toplevel)
+{
+	GString *string;
+	gchar *str;
+	
+	g_return_val_if_fail (mime_part != NULL, NULL);
+	
+	string = g_string_new ("");
+	g_mime_part_write_to_string (mime_part, toplevel, string);
+	str = string->str;
+	g_string_free (string, FALSE);
 	
 	return str;
 }
