@@ -21,9 +21,13 @@
  *
  */
 
-#include "gmime-utils.h"
-#include "gmime-part.h"
+#ifdef HAVE_CONFIG_H
 #include <config.h>
+#endif
+
+#include "gmime-utils.h"
+#include "gmime-table-private.h"
+#include "gmime-part.h"
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -33,6 +37,12 @@
 
 #define	GMIME_UUDECODE_CHAR(c) (((c) - ' ') & 077)
 
+#define GMIME_FOLD_LEN  76
+
+#ifndef HAVE_ISBLANK
+#define isblank(c) (c == ' ' || c == '\t')
+#endif
+
 static char *base64_alphabet =
 "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
 
@@ -41,25 +51,6 @@ static unsigned char tohex[16] = {
 	'8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
 };
 
-static unsigned short gmime_special_table[256] = {
-	  5,  5,  5,  5,  5,  5,  5,  5,  5,231,  7,  5,  5, 39,  5,  5,
-	  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,
-	242,448, 76,192,192,192,192,192, 76, 76,448,448, 76,448, 72,324,
-	448,448,448,448,448,448,448,448,448,448, 76, 76, 76,  4, 76, 68,
-	 76,448,448,448,448,448,448,448,448,448,448,448,448,448,448,448,
-	448,448,448,448,448,448,448,448,448,448,448,108,236,108,192, 64,
-	192,448,448,448,448,448,448,448,448,448,448,448,448,448,448,448,
-	448,448,448,448,448,448,448,448,448,448,448,192,192,192,192,  5,
-	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
-	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
-	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
-	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
-	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
-	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
-	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
-	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
-};
-
 static unsigned char gmime_base64_rank[256] = {
 	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
 	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
@@ -99,109 +90,6 @@ static unsigned char gmime_uu_rank[256] = {
 };
 
 
-enum {
-	IS_CTRL		= 1<<0,
-	IS_LWSP		= 1<<1,
-	IS_TSPECIAL	= 1<<2,
-	IS_SPECIAL	= 1<<3,
-	IS_SPACE	= 1<<4,
-	IS_DSPECIAL	= 1<<5,
-	IS_QPSAFE	= 1<<6,
-	IS_ESAFE	= 1<<7,	/* encoded word safe */
-	IS_PSAFE	= 1<<8,	/* encoded word in phrase safe */
-};
-
-#define is_ctrl(x) ((gmime_special_table[(unsigned char)(x)] & IS_CTRL) != 0)
-#define is_lwsp(x) ((gmime_special_table[(unsigned char)(x)] & IS_LWSP) != 0)
-#define is_tspecial(x) ((gmime_special_table[(unsigned char)(x)] & IS_TSPECIAL) != 0)
-#define is_type(x, t) ((gmime_special_table[(unsigned char)(x)] & (t)) != 0)
-#define is_ttoken(x) ((gmime_special_table[(unsigned char)(x)] & (IS_TSPECIAL|IS_LWSP|IS_CTRL)) == 0)
-#define is_atom(x) ((gmime_special_table[(unsigned char)(x)] & (IS_SPECIAL|IS_SPACE|IS_CTRL)) == 0)
-#define is_dtext(x) ((gmime_special_table[(unsigned char)(x)] & IS_DSPECIAL) == 0)
-#define is_fieldname(x) ((gmime_special_table[(unsigned char)(x)] & (IS_CTRL|IS_SPACE)) == 0)
-#define is_qpsafe(x) ((gmime_special_table[(unsigned char)(x)] & IS_QPSAFE) != 0)
-#define is_especial(x) ((gmime_special_table[(unsigned char)(x)] & IS_ESPECIAL) != 0)
-#define is_psafe(x) ((gmime_special_table[(unsigned char)(x)] & IS_PSAFE) != 0)
-
-#ifndef HAVE_ISBLANK
-#define isblank(c) ((c) == ' ' || (c) == '\t')
-#endif /* HAVE_ISBLANK */
-
-#define CHARS_LWSP " \t\n\r"               /* linear whitespace chars */
-#define CHARS_TSPECIAL "()<>@,;:\\\"/[]?="
-#define CHARS_SPECIAL "()<>@,;:\\\".[]"
-#define CHARS_CSPECIAL "()\\\r"	           /* not in comments */
-#define CHARS_DSPECIAL "[]\\\r \t"	   /* not in domains */
-#define CHARS_ESPECIAL "()<>@,;:\"/[]?.="  /* encoded word specials (rfc2047 5.1) */
-#define CHARS_PSPECIAL "!*+-/"             /* encoded phrase specials (rfc2047 5.3) */
-
-#ifdef BUILD_TABLE
-/* code to rebuild the gmime_special_table */
-static void
-header_remove_bits (gushort bit, guchar *vals)
-{
-	gint i;
-	
-	for (i = 0; vals[i]; i++)
-		gmime_special_table[vals[i]] &= ~bit;
-}
-
-static void
-header_init_bits (gushort bit, gushort bitcopy, gboolean remove, guchar *vals)
-{
-	gint i, len = strlen (vals);
-	
-	if (!remove) {
-		for (i = 0; i < len; i++) {
-			gmime_special_table[vals[i]] |= bit;
-		}
-		if (bitcopy) {
-			for (i = 0; i < 256; i++) {
-				if (gmime_special_table[i] & bitcopy)
-					gmime_special_table[i] |= bit;
-			}
-		}
-	} else {
-		for (i = 0; i < 256; i++)
-			gmime_special_table[i] |= bit;
-		for (i = 0; i < len; i++) {
-			gmime_special_table[vals[i]] &= ~bit;
-		}
-		if (bitcopy) {
-			for (i = 0; i < 256; i++) {
-				if (gmime_special_table[i] & bitcopy)
-					gmime_special_table[i] &= ~bit;
-			}
-		}
-	}
-}
-
-static void
-header_decode_init (void)
-{
-	gint i;
-	
-	for (i = 0; i < 256; i++) {
-		gmime_special_table[i] = 0;
-		if (i < 32)
-			gmime_special_table[i] |= IS_CTRL;
-		if ((i >= 33 && i <= 60) || (i >= 62 && i <= 126) || i == 32 || i == 9)
-			gmime_special_table[i] |= (IS_QPSAFE | IS_ESAFE);
-		if ((i >= '0' && i <= '9') || (i >= 'a' && i <= 'z') || (i >= 'A' && i <= 'Z'))
-			gmime_special_table[i] |= IS_PSAFE;
-	}
-	
-	gmime_special_table[127] |= IS_CTRL;
-	gmime_special_table[' '] |= IS_SPACE;
-	header_init_bits (IS_LWSP, 0, FALSE, CHARS_LWSP);
-	header_init_bits (IS_TSPECIAL, IS_CTRL, FALSE, CHARS_TSPECIAL);
-	header_init_bits (IS_SPECIAL, 0, FALSE, CHARS_SPECIAL);
-	header_init_bits (IS_DSPECIAL, 0, FALSE, CHARS_DSPECIAL);
-	header_remove_bits (IS_ESAFE, CHARS_ESPECIAL);
-	header_init_bits (IS_PSAFE, 0, FALSE, CHARS_PSPECIAL);
-}
-#endif /* BUILD_TABLE */
-
 /* hrm, is there a library for this shit? */
 static struct {
 	char *name;
@@ -292,22 +180,34 @@ datetok (const gchar *date)
 	return tokens;
 }
 
+#if 0
 static gint
-get_days_in_month (gint mon, gint year)
+get_days_in_month (gint month, gint year)
 {
-	switch (mon) {
-	case 1: case 3: case 5: case 7: case 8: case 10: case 12:
-		return 31;
-	case 4: case 6: case 9: case 11:
-		return 30;
+        switch (month) {
+	case 1:
+	case 3:
+	case 5:
+	case 7:
+	case 8:
+	case 10:
+	case 12:
+	        return 31;
+	case 4:
+	case 6:
+	case 9:
+	case 11:
+	        return 30;
 	case 2:
-		if ((year % 4 == 0 && year % 100 != 0) || year % 400 == 0)
-			return 29;
-		return 28;
+	        if (g_date_is_leap_year (year))
+		        return 29;
+		else
+		        return 28;
 	default:
-		return 30;
+	        return 0;
 	}
 }
+#endif
 
 static gint
 get_wday (gchar *str)
@@ -358,10 +258,10 @@ get_month (gchar *str)
 }
 
 static gint
-get_year (gchar *str)
+get_year (const gchar *str)
 {
 	gint year;
-	gchar *p;
+	const gchar * p;
 	
 	for (p = str; *p; p++)
 		if (!isdigit (*p))
@@ -370,7 +270,7 @@ get_year (gchar *str)
 	year = atoi (str);
 	
 	if (year < 100)
-		year += 1900;
+		year += (year < 70) ? 2000 : 1900;
 	
 	if (year < 1969)
 		return -1;
@@ -379,9 +279,9 @@ get_year (gchar *str)
 }
 
 static gboolean
-get_time (gchar *in, gint *hour, gint *min, gint *sec)
+get_time (const gchar *in, gint *hour, gint *min, gint *sec)
 {
-	gchar *p;
+	const gchar *p;
 	gint colons = 0;
 	gboolean digits = TRUE;
 	
@@ -392,10 +292,16 @@ get_time (gchar *in, gint *hour, gint *min, gint *sec)
 			digits = FALSE;
 	}
 	
-	if (!digits || colons != 2)
+	/* Ameol software doesn't put the seconds on the time */
+	if (!digits || (colons != 2 && colons != 1))
 		return FALSE;
 	
-	return sscanf (in, "%d:%d:%d", hour, min, sec) == 3;
+	if (colons == 2)
+		return sscanf (in, "%d:%d:%d", hour, min, sec) == 3;
+	else {
+		*sec = 0;
+		return sscanf (in, "%d:%d", hour, min) == 2;
+	}
 }
 
 static gint
@@ -508,13 +414,14 @@ parse_rfc822_date (GList *tokens, int *tzone)
 static time_t
 parse_broken_date (GList *tokens, int *tzone)
 {
+#if 0
 	GList *token;
 	struct tm tm;
 	gint hour, min, sec, n;
 	
 	if (tzone)
 		*tzone = 0;
-	
+#endif	
 	return (time_t) 0;
 }
 
@@ -552,6 +459,110 @@ g_mime_utils_header_decode_date (const gchar *in, gint *saveoffset)
 }
 
 
+/**
+ * g_mime_utils_header_fold: Fold a header.
+ * @in: input header string
+ *
+ * Returns an allocated string containing the folded header.
+ **/
+gchar *
+g_mime_utils_header_fold (const gchar *in)
+{
+	const gchar *inptr, *space;
+	gint len, outlen, i;
+	GString *out;
+	gchar *ret;
+	
+	inptr = in;
+	len = strlen (in);
+	if (len <= GMIME_FOLD_LEN)
+		return g_strdup (in);
+	
+	out = g_string_new ("");
+	outlen = 0;
+	while (*inptr) {
+		space = strchr (inptr, ' ');
+		if (space)
+			len = space - inptr + 1;
+		else
+			len = strlen (inptr);
+		
+		if (outlen + len > GMIME_FOLD_LEN) {
+			g_string_append (out, "\n\t");
+			outlen = 1;
+			
+			/* check for very long words, just cut them up */
+			while (outlen + len > GMIME_FOLD_LEN) {
+				for (i = 0; i < GMIME_FOLD_LEN - outlen; i++)
+					g_string_append_c (out, inptr[i]);
+				inptr += GMIME_FOLD_LEN - outlen;
+				len -= GMIME_FOLD_LEN - outlen;
+				g_string_append (out, "\n\t");
+				outlen = 1;
+			}
+		}
+		
+		outlen += len;
+		for (i = 0; i < len; i++) {
+			g_string_append_c (out, inptr[i]);
+		}
+		inptr += len;
+	}
+	
+	ret = out->str;
+	g_string_free (out, FALSE);
+	
+	return ret;	
+}
+
+
+/**
+ * g_mime_utils_header_printf: Format a header.
+ * @format: string format
+ * @Varargs: arguments
+ *
+ * Returns an allocated string containing the folded header specified
+ * by #format and the following arguments.
+ **/
+gchar *
+g_mime_utils_header_printf (const gchar *format, ...)
+{
+	gchar *buf, *ret;
+	va_list ap;
+	
+	va_start (ap, format);
+	buf = g_strdup_vprintf (format, ap);
+	va_end (ap);
+	
+	ret = g_mime_utils_header_fold (buf);
+	g_free (buf);
+	
+	return ret;
+}
+
+static gboolean
+need_quotes (const char *string)
+{
+	gboolean quoted = FALSE;
+	const char *inptr;
+	
+	inptr = string;
+	
+	while (*inptr) {
+		if (*inptr == '\\')
+			inptr++;
+		else if (*inptr == '"')
+			quoted = !quoted;
+		else if (!quoted && (is_tspecial (*inptr) || *inptr == '.'))
+			return TRUE;
+		
+		if (*inptr)
+			inptr++;
+	}
+	
+	return FALSE;
+}
+
 /**
  * g_mime_utils_quote_string: Quote a string.
  * @string: input string
@@ -564,18 +575,16 @@ g_mime_utils_header_decode_date (const gchar *in, gint *saveoffset)
 gchar *
 g_mime_utils_quote_string (const gchar *string)
 {
-	GString *out;
+	gboolean quote;
+	const gchar *c;
 	gchar *qstring;
-	guchar *c;
-	gboolean quote = FALSE;
+	GString *out;
 	
 	out = g_string_new ("");
+	quote = need_quotes (string);
 	
-	for (c = (guchar *) string; *c; c++) {
-		if (is_tspecial (*c))
-			quote = TRUE;
-		
-		if (*c == '"' || *c == '\\')
+	for (c = string; *c; c++) {
+		if ((*c == '"' && quote) || *c == '\\')
 			g_string_append_c (out, '\\');
 		
 		g_string_append_c (out, *c);
@@ -635,16 +644,20 @@ g_mime_utils_unquote_string (gchar *string)
 /**
  * g_mime_utils_text_is_8bit: Determine if a string contains 8bit chars
  * @text: text to check for 8bit chars
+ * @len: text length
  *
  * Returns TRUE if the text contains 8bit characters or FALSE
  * otherwise.
  **/
 gboolean
-g_mime_utils_text_is_8bit (const guchar *text)
+g_mime_utils_text_is_8bit (const guchar *text, guint len)
 {
-	guchar *c;
+	guchar *c, *inend;
 	
-	for (c = (guchar *) text; *c; c++)
+	g_return_val_if_fail (text != NULL, FALSE);
+	
+	inend = (guchar *) text + len;
+	for (c = (guchar *) text; c < inend; c++)
 		if (*c > (guchar) 127)
 			return TRUE;
 	
@@ -655,25 +668,24 @@ g_mime_utils_text_is_8bit (const guchar *text)
 /**
  * g_mime_utils_best_encoding: Determine the best encoding
  * @text: text to encode
+ * @len: text length
  *
  * Returns a GMimePartEncodingType that is determined to be the best
  * encoding type for the specified block of text. ("best" in this
  * particular case means best compression)
  **/
 GMimePartEncodingType
-g_mime_utils_best_encoding (const guchar *text)
+g_mime_utils_best_encoding (const guchar *text, guint len)
 {
-	guchar *ch;
+	guchar *ch, *inend;
 	gint count = 0;
-	gint total;
 	
-	for (ch = (guchar *) text; *ch; ch++)
+	inend = (guchar *) text + len;
+	for (ch = (guchar *) text; ch < inend; ch++)
 		if (*ch > (guchar) 127)
 			count++;
 	
-	total = (int) (ch - text);
-	
-	if ((float) count <= total * 0.17)
+	if ((float) count <= len * 0.17)
 		return GMIME_PART_ENCODING_QUOTEDPRINTABLE;
 	else
 		return GMIME_PART_ENCODING_BASE64;
@@ -696,7 +708,7 @@ quoted_decode (const guchar *in, gint len, guchar *out)
 	while (inptr < inend) {
 		c = *inptr++;
 		if (c == '=') {
-			if (inend - in >= 2) {
+			if (inend - inptr >= 2) {
 				c = toupper (*inptr++);
 				c1 = toupper (*inptr++);
 				*outptr++ = (((c >= 'A' ? c - 'A' + 10 : c - '0') & 0x0f) << 4)
@@ -725,23 +737,21 @@ quoted_decode (const guchar *in, gint len, guchar *out)
 	return -1;
 }
 
+static gboolean
+is_8bit_word_encoded (const guchar * word)
+{
+	guint len = strlen (word);
+	return len>=7 && word[0]=='=' && word[1]=='?' && word[len-2]=='?' && word[len-1]=='=';
+}
+
 static guchar *
-decode_8bit_word (const guchar *word, gboolean *was_encoded)
+decode_encoded_8bit_word (const guchar *word)
 {
 	guchar *inptr, *inend;
 	guint len;
 	
 	len = strlen (word);
 	
-	*was_encoded = TRUE;
-	
-	/* just make sure this is valid input */
-	if (len < 7 || !(word[0] == '=' && word[1] == '?' && word[len - 2] == '?' && word[len - 1] == '=')) {
-		d(fprintf (stderr, "invalid\n"));
-		*was_encoded = FALSE;
-		return g_strdup (word);
-	}
-	
 	inptr = (guchar *) word + 2;
 	inend = (guchar *) word + len - 2;
 	
@@ -796,15 +806,15 @@ gchar *
 g_mime_utils_8bit_header_decode (const guchar *in)
 {
 	GString *out, *whtspc, *word;
-	guchar *inptr;
+	const guchar *inptr;
 	guchar *decoded;
 	gboolean last_was_encoded = FALSE;
 	gboolean last_was_space = FALSE;
 	
-	out = g_string_new ("");
-	whtspc = g_string_new ("");
-	word = g_string_new ("");
-	inptr = (guchar *) in;
+	out = g_string_sized_new (256);
+	whtspc = g_string_sized_new (256);
+	word = g_string_sized_new (256);
+	inptr = in;
 	
 	while (inptr && *inptr) {
 		guchar c = *inptr++;
@@ -812,10 +822,15 @@ g_mime_utils_8bit_header_decode (const guchar *in)
 		if (isspace (c) && !last_was_space) {
 			/* we reached the end of a 'word' */
 			gboolean was_encoded;
-			guchar *dword;
+			const gchar *cword;
+			guchar *dword = NULL;
+
+			if ((was_encoded = is_8bit_word_encoded (word->str)))
+				cword = dword = decode_encoded_8bit_word (word->str);
+			else
+				cword = word->str;
 			
-			dword = decode_8bit_word (word->str, &was_encoded);
-			if (dword) {
+			if (cword) {
 				if (!(last_was_encoded && was_encoded)) {
 					/* rfc2047 states that you
                                            must ignore all whitespace
@@ -823,7 +838,7 @@ g_mime_utils_8bit_header_decode (const guchar *in)
 					g_string_append (out, whtspc->str);
 				}
 				
-				g_string_append (out, dword);
+				g_string_append (out, cword);
 				g_free (dword);
 			} else {
 				was_encoded = FALSE;
@@ -832,12 +847,9 @@ g_mime_utils_8bit_header_decode (const guchar *in)
 			}
 			
 			last_was_encoded = was_encoded;
-			
-			g_string_free (whtspc, TRUE);
-			whtspc = g_string_new ("");
-			
-			g_string_free (word, TRUE);
-			word = g_string_new ("");
+
+			g_string_truncate (whtspc, 0);
+			g_string_truncate (word, 0);
 		}
 		
 		if (isspace (c)) {
@@ -851,10 +863,15 @@ g_mime_utils_8bit_header_decode (const guchar *in)
 	
 	if (word->len || whtspc->len) {
 		gboolean was_encoded;
-		guchar *dword;
+		const guchar *cword;
+		guchar *dword = NULL;
+
+		if ((was_encoded = is_8bit_word_encoded (word->str)))
+			cword = dword = decode_encoded_8bit_word (word->str);
+		else
+			cword = word->str;
 		
-		dword = decode_8bit_word (word->str, &was_encoded);
-		if (dword) {
+		if (cword) {
 			if (!(last_was_encoded && was_encoded)) {
 				/* rfc2047 states that you
 				   must ignore all whitespace
@@ -862,7 +879,7 @@ g_mime_utils_8bit_header_decode (const guchar *in)
 				g_string_append (out, whtspc->str);
 			}
 			
-			g_string_append (out, dword);
+			g_string_append (out, cword);
 			g_free (dword);
 		} else {
 			g_string_append (out, whtspc->str);
@@ -893,9 +910,9 @@ quoted_encode (const guchar *in, gint len, guchar *out, gushort safemask)
 	
 	while (inptr < inend) {
 		c = *inptr++;
-		if (gmime_special_table[c] & safemask && c != '\t') {
-			if (c == ' ')
-				c = '_';
+		if (c == ' ') {
+			*outptr++ = '_';
+		} else if (gmime_special_table[c] & safemask) {
 			*outptr++ = c;
 		} else {
 			*outptr++ = '=';
@@ -918,7 +935,7 @@ encode_8bit_word (const guchar *word, gushort safemask, gboolean *this_was_encod
 	
 	len = strlen (word);
 	
-	switch (g_mime_utils_best_encoding (word)) {
+	switch (g_mime_utils_best_encoding (word, len)) {
 	case GMIME_PART_ENCODING_BASE64:
 		enclen = BASE64_ENCODE_LEN (len);
 		encoded = alloca (enclen);
@@ -1026,12 +1043,9 @@ g_mime_utils_8bit_header_encode (const guchar *in)
 			/* append the encoded word */
 			g_string_append (out, eword);
 			g_free (eword);
-			
-			g_string_free (whtspc, TRUE);
-			whtspc = g_string_new ("");
-			
-			g_string_free (word, TRUE);
-			word = g_string_new ("");
+
+			g_string_truncate (whtspc, 0);
+			g_string_truncate (word, 0);
 			
 			last_was_encoded = this_was_encoded;
 			is8bit = FALSE;
